from contextlib import suppress
from typing import TYPE_CHECKING, Any, TypeVar

from sqlalchemy import inspect, sql
from sqlalchemy.orm import Mapper
from sqlalchemy.sql import ColumnCollection, ColumnElement
from sqlalchemy.sql.base import ReadOnlyColumnCollection

from ixmp4.core.exceptions import ProgrammingError

if TYPE_CHECKING:
    from ixmp4.data.db.base import BaseModel

JoinType = TypeVar("JoinType", bound=sql.Select[tuple["BaseModel", ...]])


# This should not need to be Any, I think, but if I put "BaseModel" instead, various
# things like RunMetaEntry are not recognized -> this sounds like covarianve again,
# but covariant typevars are not allowed as type hints.
def is_joined(exc: sql.Select[tuple[Any, ...]], model: type["BaseModel"]) -> bool:
    """Returns `True` if `model` has been joined in `exc`."""
    for visitor in sql.visitors.iterate(exc):
        # Checking for `.join(Child)` clauses
        if visitor.__visit_name__ == "table":
            # Visitor might be of ColumnCollection or so,
            # which cannot be compared to model
            with suppress(TypeError):
                if model == visitor.entity_namespace:  # type: ignore[attr-defined]
                    return True
    return False


def get_columns(model_class: type) -> ColumnCollection[str, ColumnElement[Any]]:
    mapper: Mapper[Any] | None = inspect(model_class)
    if mapper is not None:
        return mapper.selectable.columns
    else:
        raise ProgrammingError(f"Model class `{model_class.__name__}` is not mapped.")


def get_pk_columns(
    model_class: type,
) -> ReadOnlyColumnCollection[str, ColumnElement[int]]:
    columns: ColumnCollection[str, ColumnElement[int]] = ColumnCollection()
    for col in get_columns(model_class):
        if col.primary_key:
            columns.add(col)

    return columns.as_readonly()


def get_foreign_columns(
    model_class: type,
) -> ReadOnlyColumnCollection[str, ColumnElement[int]]:
    columns: ColumnCollection[str, ColumnElement[int]] = ColumnCollection()
    for col in get_columns(model_class):
        if len(col.foreign_keys) > 0:
            columns.add(col)

    return columns.as_readonly()



from typing import ClassVar

from ixmp4 import db
from ixmp4.data import abstract, types
from ixmp4.data.db.model.model import Model
from ixmp4.data.db.optimization.equation import Equation
from ixmp4.data.db.optimization.indexset import IndexSet
from ixmp4.data.db.optimization.parameter import Parameter
from ixmp4.data.db.optimization.scalar import Scalar
from ixmp4.data.db.optimization.table import Table
from ixmp4.data.db.optimization.variable import Variable as OptimizationVariable
from ixmp4.data.db.scenario.model import Scenario

from .. import base, mixins, versions


class Run(base.BaseModel, mixins.HasUpdateInfo):
    __tablename__ = "run"
    IsLocked: ClassVar = abstract.Run.IsLocked
    NotFound: ClassVar = abstract.Run.NotFound
    NotUnique: ClassVar = abstract.Run.NotUnique
    DeletionPrevented: ClassVar = abstract.Run.DeletionPrevented

    NoDefaultVersion: ClassVar = abstract.Run.NoDefaultVersion

    __table_args__ = (db.UniqueConstraint("model__id", "scenario__id", "version"),)

    model__id: types.Integer = db.Column(
        db.ForeignKey("model.id"), nullable=False, index=True
    )
    model: types.Mapped[Model] = db.relationship(
        "Model",
        backref="run",
        foreign_keys=[model__id],
    )

    scenario__id: types.Integer = db.Column(
        db.ForeignKey("scenario.id"),
        nullable=False,
        index=True,
    )
    scenario: types.Mapped[Scenario] = db.relationship(
        "Scenario",
        backref="run",
        foreign_keys=[scenario__id],
    )

    equations: types.Mapped[list["Equation"]] = db.relationship()
    indexsets: types.Mapped[list["IndexSet"]] = db.relationship()
    parameters: types.Mapped[list["Parameter"]] = db.relationship()
    scalars: types.Mapped[list["Scalar"]] = db.relationship()
    tables: types.Mapped[list["Table"]] = db.relationship()
    variables: types.Mapped[list["OptimizationVariable"]] = db.relationship()

    version: types.Integer = db.Column(db.Integer, nullable=False)
    is_default: types.Boolean = db.Column(db.Boolean, default=False, nullable=False)
    lock_transaction: types.Mapped[int | None] = db.Column(nullable=True, index=True)


class RunVersion(versions.DefaultVersionModel):
    __tablename__ = "run_version"
    model__id: types.Integer = db.Column(nullable=False, index=True)
    scenario__id: types.Integer = db.Column(nullable=False, index=True)
    version: types.Integer = db.Column(nullable=False)
    is_default: types.Boolean = db.Column(nullable=False)
    lock_transaction: types.Integer = db.Column(nullable=True, index=True)

    created_at: types.DateTime = db.Column(nullable=True)
    created_by: types.Username
    updated_at: types.DateTime = db.Column(nullable=True)
    updated_by: types.Username


version_triggers = versions.PostgresVersionTriggers(Run.__table__, RunVersion.__table__)



from typing import Any, TypeVar

from ixmp4 import db

from .. import base
from .model import DefaultVersionModel, Operation

VModelType = TypeVar("VModelType", bound=DefaultVersionModel)


class VersionRepository(
    base.Retriever[VModelType],
    base.Enumerator[VModelType],
):
    model_class: type[VModelType]

    def where_matches_kwargs(
        self, exc: db.sql.Select[Any], **kwargs: Any
    ) -> db.sql.Select[Any]:
        for key in kwargs:
            columns = db.utils.get_columns(self.model_class)
            if key in columns:
                exc = exc.where(columns[key] == kwargs[key])
        return exc

    def where_valid_at_transaction(
        self,
        exc: db.sql.Select[Any],
        transaction__id: int,
        cls: type[DefaultVersionModel] | None = None,
    ) -> db.sql.Select[Any]:
        if cls is None:
            cls = self.model_class
        return exc.where(
            db.and_(
                cls.transaction_id <= transaction__id,
                cls.operation_type != Operation.DELETE,
                db.or_(
                    cls.end_transaction_id > transaction__id,
                    cls.end_transaction_id == db.null(),
                ),
            )
        )

    def select(
        self, transaction__id: int | None = None, **kwargs: Any
    ) -> db.sql.Select[Any]:
        exc = db.select(self.model_class)

        if transaction__id is not None:
            exc = self.where_valid_at_transaction(exc, transaction__id)

        exc = self.where_matches_kwargs(exc, **kwargs)

        return exc.order_by(self.model_class.id.asc())



from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, cast

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import Unpack

if TYPE_CHECKING:
    from ixmp4.data.backend.db import SqlAlchemyBackend


import logging

import pandas as pd

from ixmp4 import db
from ixmp4.core.exceptions import OptimizationItemUsageError
from ixmp4.data import types
from ixmp4.data.abstract import optimization as abstract
from ixmp4.data.auth.decorators import guard
from ixmp4.data.db.unit import Unit

from .. import base
from .docs import ParameterDocsRepository
from .model import Parameter, ParameterIndexsetAssociation

logger = logging.getLogger(__name__)


class ParameterRepository(
    base.Creator[Parameter],
    base.Deleter[Parameter],
    base.Retriever[Parameter],
    base.Enumerator[Parameter],
    abstract.ParameterRepository,
):
    model_class = Parameter

    UsageError = OptimizationItemUsageError

    def __init__(self, *args: "SqlAlchemyBackend") -> None:
        super().__init__(*args)
        self.docs = ParameterDocsRepository(*args)

        from .filter import OptimizationParameterFilter

        self.filter_class = OptimizationParameterFilter

    def add(
        self,
        run_id: int,
        name: str,
        constrained_to_indexsets: list[str],
        column_names: list[str] | None = None,
    ) -> Parameter:
        parameter = Parameter(name=name, run__id=run_id)

        indexsets = {
            indexset: self.backend.optimization.indexsets.get(
                run_id=run_id, name=indexset
            )
            for indexset in set(constrained_to_indexsets)
        }
        for i in range(len(constrained_to_indexsets)):
            _ = ParameterIndexsetAssociation(
                parameter=parameter,
                indexset=indexsets[constrained_to_indexsets[i]],
                column_name=column_names[i] if column_names else None,
            )

        parameter.set_creation_info(auth_context=self.backend.auth_context)
        self.session.add(parameter)

        return parameter

    @guard("view")
    def get(self, run_id: int, name: str) -> Parameter:
        exc = db.select(Parameter).where(
            (Parameter.name == name) & (Parameter.run__id == run_id)
        )
        try:
            return self.session.execute(exc).scalar_one()
        except db.NoResultFound:
            raise Parameter.NotFound

    @guard("view")
    def get_by_id(self, id: int) -> Parameter:
        obj = self.session.get(self.model_class, id)

        if obj is None:
            raise Parameter.NotFound(id=id)

        return obj

    @guard("edit")
    def create(
        self,
        run_id: int,
        name: str,
        constrained_to_indexsets: list[str],
        column_names: list[str] | None = None,
    ) -> Parameter:
        if column_names and len(column_names) != len(constrained_to_indexsets):
            raise self.UsageError(
                f"While processing Parameter {name}: \n"
                "`constrained_to_indexsets` and `column_names` not equal in length! "
                "Please provide the same number of entries for both!"
            )

        if column_names and len(column_names) != len(set(column_names)):
            raise self.UsageError(
                f"While processing Parameter {name}: \n"
                "The given `column_names` are not unique!"
            )

        return super().create(
            run_id=run_id,
            name=name,
            constrained_to_indexsets=constrained_to_indexsets,
            column_names=column_names,
        )

    @guard("edit")
    def delete(self, id: int) -> None:
        super().delete(id=id)

    @guard("view")
    def list(self, **kwargs: Unpack["base.EnumerateKwargs"]) -> Iterable[Parameter]:
        return super().list(**kwargs)

    @guard("view")
    def tabulate(self, **kwargs: Unpack["base.EnumerateKwargs"]) -> pd.DataFrame:
        return super().tabulate(**kwargs)

    @guard("edit")
    def add_data(self, id: int, data: dict[str, Any] | pd.DataFrame) -> None:
        if isinstance(data, dict):
            try:
                data = pd.DataFrame.from_dict(data=data)
            except ValueError as e:
                raise Parameter.DataInvalid(str(e)) from e

        if data.empty:
            return  # nothing to do

        parameter = self.get_by_id(id=id)

        missing_columns = set(["values", "units"]) - set(data.columns)
        if missing_columns:
            raise OptimizationItemUsageError(
                "Parameter.data must include the column(s): "
                f"{', '.join(missing_columns)}!"
            )

        # Can use a set for now, need full column if we care about order
        for unit_name in set(data["units"]):
            try:
                self.backend.units.get(name=unit_name)
            except Unit.NotFound as e:
                # TODO Add a helpful hint on how to check defined Units
                raise Unit.NotFound(
                    message=f"'{unit_name}' is not defined for this Platform!"
                ) from e

        index_list = parameter.column_names or parameter.indexset_names
        existing_data = pd.DataFrame(parameter.data)
        if not existing_data.empty:
            existing_data.set_index(index_list, inplace=True)

        parameter.data = cast(
            types.JsonDict,
            (
                data.set_index(index_list).combine_first(existing_data).reset_index()
            ).to_dict(orient="list"),
        )

        self.session.commit()

    @guard("edit")
    def remove_data(self, id: int, data: dict[str, Any] | pd.DataFrame) -> None:
        if isinstance(data, dict):
            data = pd.DataFrame.from_dict(data=data)

        if data.empty:
            return

        parameter = self.get_by_id(id=id)
        index_list = parameter.column_names or parameter.indexset_names
        existing_data = pd.DataFrame(parameter.data)
        if not existing_data.empty:
            existing_data.set_index(index_list, inplace=True)

        # This is the only kind of validation we do for removal data
        try:
            data.set_index(index_list, inplace=True)
        except KeyError as e:
            logger.error(
                f"Data to be removed must include {index_list} as keys/columns, but "
                f"{[name for name in data.columns]} were provided."
            )
            raise OptimizationItemUsageError(
                "The data to be removed must specify one or more complete indices to "
                "remove associated units and values!"
            ) from e

        remaining_data = existing_data[~existing_data.index.isin(data.index)]
        if not remaining_data.index.empty:
            remaining_data.reset_index(inplace=True)

        parameter.data = cast(types.JsonDict, remaining_data.to_dict(orient="list"))



from datetime import datetime
from typing import TYPE_CHECKING, Any, ClassVar, cast

if TYPE_CHECKING:
    from ixmp4.core.run import Run

    from . import InitKwargs

import pandas as pd

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import Unpack

from ixmp4.data.abstract import Docs as DocsModel
from ixmp4.data.abstract import Parameter as ParameterModel
from ixmp4.data.abstract import Unit

from .base import (
    Creator,
    Deleter,
    Lister,
    OptimizationBaseModelFacade,
    Retriever,
    Tabulator,
)


class Parameter(OptimizationBaseModelFacade):
    _model: ParameterModel
    NotFound: ClassVar = ParameterModel.NotFound
    NotUnique: ClassVar = ParameterModel.NotUnique

    @property
    def id(self) -> int:
        return self._model.id

    @property
    def name(self) -> str:
        return self._model.name

    @property
    def run_id(self) -> int:
        return self._model.run__id

    @property
    def data(self) -> dict[str, list[float] | list[int] | list[str]]:
        return self._model.data

    def add(self, data: dict[str, Any] | pd.DataFrame) -> None:
        """Adds data to the Parameter."""
        self._run.require_lock()
        self.backend.optimization.parameters.add_data(id=self._model.id, data=data)
        self._model = self.backend.optimization.parameters.get(
            run_id=self._model.run__id, name=self._model.name
        )

    def remove(self, data: dict[str, Any] | pd.DataFrame) -> None:
        """Removes data from the Parameter.

        The data must specify all indexed columns. All other keys/columns are ignored.
        """
        self._run.require_lock()
        self.backend.optimization.parameters.remove_data(id=self._model.id, data=data)
        self._model = self.backend.optimization.parameters.get(
            run_id=self._model.run__id, name=self._model.name
        )

    @property
    def values(self) -> list[float]:
        return cast(list[float], self._model.data.get("values", []))

    @property
    def units(self) -> list[Unit]:
        return cast(list[Unit], self._model.data.get("units", []))

    @property
    def indexset_names(self) -> list[str]:
        return self._model.indexset_names

    @property
    def column_names(self) -> list[str] | None:
        return self._model.column_names

    @property
    def created_at(self) -> datetime | None:
        return self._model.created_at

    @property
    def created_by(self) -> str | None:
        return self._model.created_by

    @property
    def docs(self) -> str | None:
        try:
            return self.backend.optimization.parameters.docs.get(self.id).description
        except DocsModel.NotFound:
            return None

    @docs.setter
    def docs(self, description: str | None) -> None:
        if description is None:
            self.backend.optimization.parameters.docs.delete(self.id)
        else:
            self.backend.optimization.parameters.docs.set(self.id, description)

    @docs.deleter
    def docs(self) -> None:
        try:
            self.backend.optimization.parameters.docs.delete(self.id)
        # TODO: silently failing
        except DocsModel.NotFound:
            return None

    def __str__(self) -> str:
        return f"<Parameter {self.id} name={self.name}>"


class ParameterRepository(
    Creator[Parameter, ParameterModel],
    Deleter[Parameter, ParameterModel],
    Retriever[Parameter, ParameterModel],
    Lister[Parameter, ParameterModel],
    Tabulator[Parameter, ParameterModel],
):
    def __init__(self, _run: "Run", **kwargs: Unpack["InitKwargs"]) -> None:
        super().__init__(_run=_run, **kwargs)
        self._backend_repository = self.backend.optimization.parameters
        self._model_type = Parameter

    def create(
        self,
        name: str,
        constrained_to_indexsets: list[str],
        column_names: list[str] | None = None,
    ) -> Parameter:
        return super().create(
            name=name,
            constrained_to_indexsets=constrained_to_indexsets,
            column_names=column_names,
        )
